/* 
various utilities for writing stuff to buffers.
these buffers can then be written to files or sent over the network.

NOTE:
wbuf = write buffer
rbuf = read buffer
*/
static WriteBuffer mkwbuffer(u8 *buf, size_t n) {
	WriteBuffer b = {0};
	b.cur = buf;
	b.original_size = n;
	b.end = buf + n;
	return b;
}

static void wbuf_err(WriteBuffer *b, char const *fmt, ...) {
	if (b->error) {
		// it's okay; we already got an error.
	} else {
		char str[2048];
		va_list args;
		va_start(args, fmt);
		vsprintf(str, fmt, args);
		va_end(args);
		str_cpy(b->error_message, sizeof b->error_message, str);
		b->error = true;
		b->cur = b->end;
	}
}

static void rbuf_err(ReadBuffer *b, char const *fmt, ...) {
	if (b->error) {
		// it's okay; we already got an error.
	} else {
		char str[2048];
		va_list args;
		va_start(args, fmt);
		vsprintf(str, fmt, args);
		va_end(args);
		str_cpy(b->error_message, sizeof b->error_message, str);
		b->error = true;
		b->cur = b->end;
	}
}

static size_t buf_bytes_written(WriteBuffer *b) {
	return b->original_size - (size_t)(b->end - b->cur);
}

static bool wbuf_has_bytes(WriteBuffer *b, size_t n) {
	return n <= (size_t)(b->end - b->cur);
}

static void buf_write_u8(WriteBuffer *b, u8 x) {
	if (wbuf_has_bytes(b, 1)) {
		*b->cur++ = x;
	} else {
		wbuf_err(b, "Ran out of space (writing u8)");
	}
}

static void buf_write_i8(WriteBuffer *b, i8 x) {
	buf_write_u8(b, (u8)x);
}

static void buf_write_u16(WriteBuffer *b, u16 x) {
	if (wbuf_has_bytes(b, 2)) {
		*b->cur++ = (u8)(x & 0xFF);
		*b->cur++ = (u8)((x >> 8) & 0xFF);
	} else {
		wbuf_err(b, "Ran out of space (writing u16)");
	}
}

static void buf_write_i16(WriteBuffer *b, i16 x) {
	buf_write_u16(b, (u16)x);
}

static void buf_write_u32(WriteBuffer *b, u32 x) {
	if (wbuf_has_bytes(b, 4)) {
		*b->cur++ = (u8)(x & 0xFF);
		*b->cur++ = (u8)((x >> 8) & 0xFF);
		*b->cur++ = (u8)((x >> 16) & 0xFF);
		*b->cur++ = (u8)((x >> 24) & 0xFF);
	} else {
		wbuf_err(b, "Ran out of space (writing u32)");
	}
}

static void buf_write_i32(WriteBuffer *b, i32 x) {
	buf_write_u32(b, (u32)x);
}

static void buf_write_hex_pos(WriteBuffer *b, HexPos pos) {
	buf_write_u16(b, pos.x);
	buf_write_u16(b, pos.y);
}

static void buf_write_u8s(WriteBuffer *b, u8 *u8s, size_t count) {
	if (wbuf_has_bytes(b, count)) {
		memcpy(b->cur, u8s, count);
		b->cur += count;
	} else {
		wbuf_err(b, "Ran out of space (writing %lu u8s)", (ulong)count);
	}
}

static void buf_write_cstr(WriteBuffer *b, char *s) {
	size_t count = strlen(s)+1;
	buf_write_u8s(b, (u8 *)s, count);
}

static ReadBuffer mkrbuffer(u8 const *buf, size_t n) {
	ReadBuffer b = {0};
	b.cur = buf;
	b.end = buf + n;
	return b;
}

static bool rbuf_has_bytes(ReadBuffer *b, size_t n) {
	return n <= (size_t)(b->end - b->cur);
}

static u8 buf_read_u8(ReadBuffer *b) {
	if (rbuf_has_bytes(b, 1)) {
		return *b->cur++;
	} else {
		rbuf_err(b, "Nothing left to read (reading u8).");
		return 0;
	}
}

// this code technically isn't standard compliant, because it relies on two's complement
// but seriously, no computers use anything else nowadays

static i8 buf_read_i8(ReadBuffer *b) {
	return (i8)buf_read_u8(b);
}

static u16 buf_read_u16(ReadBuffer *b) {
	if (rbuf_has_bytes(b, 2)) {
		u8 byte1 = *b->cur++;
		u8 byte2 = *b->cur++;
		return (u16)(byte1 | ((u16)byte2 << 8));
	} else {
		rbuf_err(b, "Nothing left to read (reading u16).");
		return 0;
	}
}

static i16 buf_read_i16(ReadBuffer *b) {
	return (i16)buf_read_u16(b);
}

static u32 buf_read_u32(ReadBuffer *b) {
	if (rbuf_has_bytes(b, 4)) {
		u8 byte1 = *b->cur++;
		u8 byte2 = *b->cur++;
		u8 byte3 = *b->cur++;
		u8 byte4 = *b->cur++;
		return byte1 | ((u32)byte2 << 8) | ((u32)byte3 << 16) | ((u32)byte4 << 24);
	} else {
		rbuf_err(b, "Nothing left to read (reading u32).");
		return 0;
	}
}

static i32 buf_read_i32(ReadBuffer *b) {
	return (i32)buf_read_u32(b);
}

static HexPos buf_read_hex_pos(ReadBuffer *b, u16 grid_w, u16 grid_h) {
	HexPos h = {0};
	u16 x = buf_read_u16(b);
	u16 y = buf_read_u16(b);
	if (x > grid_w || y > grid_h) {
		rbuf_err(b, "Invalid hex position. Got (%u, %u), but the grid size is %ux%u.", x, y, grid_w, grid_h);
		b->error = true;
		return h;
	}
	h.x = x;
	h.y = y;
	return h;
}

static void buf_read_u8s(ReadBuffer *b, u8 *u8s, size_t count) {
	if (rbuf_has_bytes(b, count)) {
		memcpy(u8s, b->cur, count);
		b->cur += count;
	} else {
		rbuf_err(b, "Nothing left to read (reading %lu u8s).", (ulong)count);
	}
}

static char const *buf_read_cstr(ReadBuffer *b) {
	char const *start = (char const *)b->cur;
	u8 const *end = b->end;
	for (; b->cur < end; ++b->cur) {
		if (*b->cur == 0) break;
	}
	if (b->cur == end) {
		b->error = true;
		return NULL;
	}
	return start;
}

static void buf_write_actions(WriteBuffer *b, Action const *actions, size_t nactions) {
	Action const *act, *end;
	buf_write_u32(b, (u32)nactions);
	for (act = actions, end = act + nactions; act != end; ++act) {
		buf_write_u8(b, (u8)act->kind);
		buf_write_hex_pos(b, act->hex1);
		buf_write_hex_pos(b, act->hex2);
	}
}

static u32 buf_read_actions(ReadBuffer *b, Game *game, Action *actions) {
	Action *act, *end;
	u32 nactions = buf_read_u32(b);
	u32 points_left = game->starting_points;
	if (nactions > ACTIONS_MAX) {
		rbuf_err(b, "Too many actions (%lu).", (ulong)nactions);
		return 0;
	}
	for (act = actions, end = act + nactions; act != end; ++act) {
		act->kind = buf_read_u8(b);
		act->hex1 = buf_read_hex_pos(b, game->width, game->height);
		act->hex2 = buf_read_hex_pos(b, game->width, game->height);

		{
			u32 cost = action_cost(game, act);
			if (points_left < cost) {
				rbuf_err(b, "Other player did more actions than they can afford!");
				return 0;
			}
			points_left -= cost;
		}

		if (!action_is_legal(game, act, actions, (u32)(act - actions))) {
			rbuf_err(b, "Illegal action (action #%ld).", (long)(act - actions));
			return 0;
		}
	}
	return nactions;
}

/* 
controlled_player needn't be game->controlled_player.
if we're writing the setup for another player to follow, they should have a 
different controlled_player
*/
static void buf_write_game_setup(WriteBuffer *b, Game *game, u8 controlled_player) {
	buf_write_u32(b, MAP_VERSION);
	buf_write_u8(b, game->nplayers);
	buf_write_u8(b, controlled_player);
	buf_write_u16(b, game->width);
	buf_write_u16(b, game->height);
	buf_write_u16(b, game->starting_points);
	buf_write_i16(b, game->starting_points_increment);

	{ // tile data 
		size_t area = (size_t)game->width * (size_t)game->height;
		buf_write_u8s(b, game->tiles, area);
	}

	{ // pion data 
		Pion *pions = game->pions, *m, *end;
		size_t npions = game->npions;
		buf_write_u32(b, (u32)npions);
		for (m = pions, end = m + npions; m != end; ++m) {
			buf_write_hex_pos(b, m->pos);
			buf_write_u8(b, m->player);
		}
	}
}

/*
needs access to all of state even though the game is just &state->game.
*/
static void buf_read_game_setup(ReadBuffer *b, State *state) {
	u32 version = buf_read_u32(b);
	u8 nplayers = buf_read_u8(b);
	u8 controlled_player = buf_read_u8(b);
	u16 grid_width  = buf_read_u16(b);
	u16 grid_height = buf_read_u16(b);
	u16 starting_points = buf_read_u16(b);
	i16 starting_points_increment = buf_read_i16(b);

	size_t grid_area = (size_t)grid_width * (size_t)grid_height;
	
	Game *game = &state->game;
	memset(game, 0, sizeof *game);

	if (version != MAP_VERSION) {
		rbuf_err(b, "Version mismatch (%lu, expected %lu).", (ulong)version, (ulong)MAP_VERSION);
		return;
	}
	
	if (nplayers < 2 || nplayers > PLAYERS_MAX) {
		rbuf_err(b, "Invalid number of players (%u)!", nplayers);
		return;
	}
	game->nplayers = nplayers;

	if (controlled_player >= nplayers) {
		rbuf_err(b, "Invalid controlled player (%u)", controlled_player);
		return;
	}
	game->controlled_player = controlled_player;

	if (grid_width == 0 || grid_width > GRID_WIDTH_MAX) {
		rbuf_err(b, "Invalid grid width (%u)", grid_width);
		return;
	}
	game->width = grid_width;
	if (grid_height == 0 || grid_height > GRID_HEIGHT_MAX) {
		rbuf_err(b, "Invalid grid height (%u)", grid_height);
		return;
	}
	game->height = grid_height;

	if (starting_points > STARTING_POINTS_MAX) {
		rbuf_err(b, "Starting points too high (%u).", starting_points);
		return;
	}
	game->starting_points = starting_points;
	if (starting_points_increment < -STARTING_POINTS_MAX || starting_points_increment > +STARTING_POINTS_MAX) {
		rbuf_err(b, "Invalid starting points increment (%d).", starting_points);
		return;
	}
	game->starting_points_increment = starting_points_increment;
	

	if (!b->error) {
		u8 *tile, *end;
		buf_read_u8s(b, game->tiles, grid_area);
		for (tile = game->tiles, end = tile + grid_area; tile != end; ++tile) {
			if (*tile >= TL_COUNT) {
				rbuf_err(b, "Invalid tile type: %u.", *tile);
				memset(game->tiles, 0, sizeof game->tiles); // make sure we don't use any invalid values
				return;
			}
		}
	}
	
	if (!b->error) { // pion data 
		u32 npions = buf_read_u32(b);
		Pion *pions = game->pions, *p, *end;
		if (npions > PIONS_MAX) {
			rbuf_err(b, "Too many pions!");
			return;
		}
		game->npions = npions;
		for (p = pions, end = p + npions; p != end; ++p) {
			p->pos = buf_read_hex_pos(b, grid_width, grid_height);
			p->player = buf_read_u8(b);
			if (p->player >= nplayers) {
				rbuf_err(b, "Invalid pion player number!");
				return;
			}
		}
		{
			u32 mark = tmp_push(state);
			bool *grid = tmp_alloc_arr(state, bool, grid_area);
			for (p = pions, end = p + npions; p != end; ++p) {
				u32 i = hex_idx(game, p->pos);
				if (grid[i]) {
					rbuf_err(b, "Two pions have the same position, (%u, %u).", p->pos.x, p->pos.y);
					tmp_pop(state, mark);
					return;
				}
				grid[i] = true;
			}
			tmp_pop(state, mark);
		}
	}

	start_turn(game);
}

