// MACROS 

/* 
	There are two versions here: the network version, and the map version.
	Increment NET_VERSION whenever a change is made that breaks backwards or forwards compatibility.
	Increment MAP_VERSION whenever a change is made that breaks backwards compatibility.
	For example, if a new tile type is added, you don't need to increment MAP_VERSION, because
	old maps will still load correctly, but you do need to increment NET_VERSION, because
	if someone with the new tile type plays a game against someone without it, there'll be problems.
	This does mean that a version mismatch won't be detected when you load a map created with a newer
	version in an older version, but people don't do that much.
*/
#define NET_VERSION 1
#define MAP_VERSION 1

#define STRINGIFY2(x) #x
#define STRINGIFY(x) STRINGIFY2(x)

#define BASIC_MOVE_POINT_COST 50
#define BASIC_ATTACK_POINT_COST 50

#define MAP_NAME_MAX_LEN 15
#define MAP_NAME_BUFSZ (MAP_NAME_MAX_LEN+1)
#define MAP_MAX_FILE_SIZE 100000 // maximum size of a map file in bytes

#define PLAYER_NONE U8_MAX
#define GRID_WIDTH_MAX 100
#define GRID_HEIGHT_MAX 100
#define GRID_AREA_MAX (GRID_WIDTH_MAX * GRID_HEIGHT_MAX)
#define PIONS_MAX GRID_AREA_MAX
#define PLAYERS_MAX 2
#define ACTIONS_MAX 100
#define TMP_MEM_BYTES (4L<<20)

#define DFL_STARTING_POINTS 150
#define DFL_POINTS_INC 25
#define STARTING_POINTS_MAX 10000 // this should be less than I16_MAX (starting_points_increment is compared against it)

#define NET_READY "READY"
#define NET_READY_SIZE strlen(NET_READY)
#define BUG_EMAIL "pommicket@pommicket.com"

// use strict enums in debug C builds, for -Wswitch (we do stuff like use the increment operator on enums,
// which isn't legal in C++)
#if DEBUG && !defined __cplusplus
#define ENUM(enumeration, type) enumeration
#else
#define ENUM(enumeration, type) type
#endif

#if DEBUG && defined __GNUC__
#define debug_log(...) printf("[%s:%d] ", __FILE__, __LINE__), printf(__VA_ARGS__), fflush(stdout)
#else
static void debug_log(char const *fmt, ...) {
	(void)fmt;
}
#endif

// CONSTANTS 
static float const main_text_sizes[] = {0.03f, 0.04f, 0.05f};
static float const big_text_size = 0.07f;

static v4 const player_colors[PLAYERS_MAX] = {
	{0.5f, 0.4f, 0.7f, 1.0f},
	{0.5f, 0.7f, 0.4f, 1.0f}
};

static v4 const error_message_color = {
	1.0f, 0.0f, 0.0f, 1.0f
};

static v4 const points_color = {
	1.0f, 1.0f, 0.6f, 1.0f
};

// UTIL 

typedef enum {
	DIR_LEFT,
	DIR_UP,
	DIR_RIGHT,
	DIR_DOWN
} Direction;

// RENDERING 
/*
	right now everything works with OpenGL 2.0 . If this changes, make sure you update main-sdl.c
	don't use GLchar ; it's not defined on windows
*/
typedef void (APIENTRY *GLAttachShader)(GLuint program, GLuint shader);
typedef void (APIENTRY *GLCompileShader)(GLuint shader);
typedef GLuint (APIENTRY *GLCreateProgram)(void);
typedef GLuint (APIENTRY *GLCreateShader)(GLenum shaderType);
typedef void (APIENTRY *GLDeleteProgram)(GLuint program);
typedef void (APIENTRY *GLDeleteShader)(GLuint shader);
typedef void (APIENTRY *GLDeleteTextures)(GLsizei n, const GLuint * textures);
typedef void (APIENTRY *GLGenerateMipmap)(GLenum target);
typedef GLint (APIENTRY *GLGetAttribLocation)(GLuint program, const char *name);
typedef void (APIENTRY *GLGetProgramInfoLog)(GLuint program, GLsizei maxLength, GLsizei *length, char *infoLog);
typedef void (APIENTRY *GLGetProgramiv)(GLuint program, GLenum pname, GLint *params);
typedef void (APIENTRY *GLGetShaderInfoLog)(GLuint shader, GLsizei maxLength, GLsizei *length, char *infoLog);
typedef void (APIENTRY *GLGetShaderiv)(GLuint shader, GLenum pname, GLint *params);
typedef GLint (APIENTRY *GLGetUniformLocation)(GLuint program, char const *name);
typedef void (APIENTRY *GLLinkProgram)(GLuint program);
typedef void (APIENTRY *GLShaderSource)(GLuint shader, GLsizei count, const char *const *string, const GLint *length);
typedef void (APIENTRY *GLUniform1f)(GLint location, float v0);
typedef void (APIENTRY *GLUniformMatrix4fv)(GLint location, GLsizei count, GLboolean transpose, const GLfloat *value);
typedef void (APIENTRY *GLUseProgram)(GLuint program);
typedef void (APIENTRY *GLVertexAttrib1f)(GLuint index, GLfloat v0);
typedef void (APIENTRY *GLVertexAttrib2f)(GLuint index, GLfloat v0, GLfloat v1);
typedef void (APIENTRY *GLVertexAttrib3f)(GLuint index, GLfloat v0, GLfloat v1, GLfloat v2);
typedef void (APIENTRY *GLVertexAttrib4f)(GLuint index, GLfloat v0, GLfloat v1, GLfloat v2, GLfloat v3);

typedef struct {
	GLAttachShader AttachShader;
	GLCompileShader CompileShader;
	GLCreateProgram CreateProgram;
	GLCreateShader CreateShader;
	GLDeleteProgram DeleteProgram;
	GLDeleteShader DeleteShader;
	GLDeleteTextures DeleteTextures;
	GLGenerateMipmap GenerateMipmap;
	GLGetAttribLocation GetAttribLocation;
	GLGetProgramInfoLog GetProgramInfoLog;
	GLGetProgramiv GetProgramiv;
	GLGetShaderInfoLog GetShaderInfoLog;
	GLGetShaderiv GetShaderiv;
	GLGetUniformLocation GetUniformLocation;
	GLLinkProgram LinkProgram;
	GLShaderSource ShaderSource;
	GLUniform1f Uniform1f;
	GLUniformMatrix4fv UniformMatrix4fv;
	GLUseProgram UseProgram;
	GLVertexAttrib1f VertexAttrib1f;
	GLVertexAttrib2f VertexAttrib2f;
	GLVertexAttrib3f VertexAttrib3f;
	GLVertexAttrib4f VertexAttrib4f;
} GLProcs;

typedef struct {
	GLuint tex;
	int width, height;
#if DEBUG
	char filename[32];
	time_t last_mtime;
#endif
} Texture;

typedef struct {
	u16 pix_w, pix_h;
	float x, y, w, h;
} Sprite;

typedef GLint UniformLocation;
typedef GLuint AttributeLocation;

typedef struct {
	GLuint program;
	UniformLocation transform;
	AttributeLocation bias;
} HexProgram;

typedef struct {
	GLuint program;
	UniformLocation transform;
	AttributeLocation x_pos;
} TitleProgram;

typedef struct {
	u16 x;
	u16 y;
} HexPos;

static HexPos hex_pos(uint x, uint y) {
	HexPos h;
	h.x = (u16)x;
	h.y = (u16)y;
	return h;
}

typedef enum {
	SPR_NONE,
	SPR_PION,
	SPR_ARROW,
	SPR_ATTACK,
	SPR_CBOX,
	SPR_CBOX_CHECKED,
	SPR_ARROWHEAD
} SpriteID;

typedef union {
	double a;
	long b;
	void *c;
	void (*d)(void);
} MaxAlign;

typedef struct {
	float res_y; // y resolution this font was created for 
	float char_height, char_height_gl;
	GLuint texture;
	int tex_width, tex_height;
	u8 const *data;
	stbtt_bakedchar char_data[96];
} TextFont; // can't just be Font; that name is taken by Xlib 

// UI 
// a full screen option menu 
typedef struct {
	char const* const *options;
	int noptions;
	int sel;
} SelectMenu;

typedef struct {
	char const *text;
	TextFont *font;
	SpriteID sprite; // if text is NULL, sprite is used 
	v2 size; // must be set if sprite is used 
	Direction direction;
	v2 pos;
	v4 color; // defaults to 1,1,1 
	bool disabled;
	bool center;
} Button;

typedef struct {
	TextFont *font;
	v2 pos;
	bool checked;
} Checkbox;

typedef struct {
	int (*char_predicate)(int); // which characters are acceptable? (defaults to isprint) 
	TextFont *font; // if not NULL, font to use for rendering text 
	char const *placeholder; // if not NULL, placeholder text, to be shown when the text input is empty and isn't in focus 
	float width; // if not 0, width of input box 
	v2 pos;
	v4 color;
	u16 max_len; // if not 0, maximum length of text 
	bool center;
	bool focus;
#define TEXT_INPUT_BUFSZ 64
	char text[TEXT_INPUT_BUFSZ];
} TextInput;

// COMMUNICATION 

typedef enum {
	NET_CONNECTING,
	NET_VERIFYING_CLIENT,
	NET_TRANSFERRING_GAME_DETAILS
} NetworkStage;

typedef struct {
	u8 *cur; // current byte we should write to 
	u8 *end;
	size_t original_size;
	bool error; // set to true if we ran out of space 
	char error_message[128];
} WriteBuffer;

typedef struct {
	u8 const *cur;
	u8 const *end;
	bool error;
	char error_message[128];
} ReadBuffer;

#define COMM_MAX_BYTES 65536
typedef struct {
	bool is_receiving, is_sending;
	u8 nbytes_of_size_received; // how many bytes of recv_size_buf have we received (4 = fully received size) 
	u8 recv_size_buf[4]; // bytes of the size, little endian 
	size_t recv_size; // number of bytes to receive 
	size_t nrecvd; // number of bytes received so far 
		
	size_t send_size; // number of bytes to send 
	size_t nsent; // number of bytes sent so far 
	u8 send_data[COMM_MAX_BYTES+4]; // first 4 bytes are the size 
	u8 received[COMM_MAX_BYTES];
	u8 *sending; // set to send_data + 4 
} Communication;

// GAME 
typedef enum {
	ACT_NONE,
	ACT_MOVE,
	ACT_ATTACK
} ActionKind;

typedef struct {
	ENUM(ActionKind, u8) kind;
	/*
		NOTE: set these hex positions to HEX_POS_INVALID if you aren't using them (middle mouse button handling
		depends on it right now, and it's probably a good idea in general)
	*/
	HexPos hex1;
	HexPos hex2;
} Action;

typedef struct {
	HexPos pos;
	HexPos new_pos;
	u8 player;
	bool will_die;
} Pion;

// tile types 
enum {
	TL_NONE,
	TL_REG,
	TL_COUNT
};

typedef Action ActionArray[ACTIONS_MAX];

typedef enum {
	GAME_HOTSEAT,
	GAME_LAN
} GameKind;

typedef enum {
	GS_PLAYING,
	GS_WAITING_FOR_OTHER_PLAYERS,
	GS_TRANSFERRING_ACTIONS,
	GS_SIMULATING_TICKS,
	GS_TRYING_TO_QUIT,
	GS_DONE
} GameStatus;

typedef enum {
	SCR_MAIN_MENU = 0,
	SCR_SETTINGS_MENU,
	SCR_SETTINGS_VIDEO,
	SCR_SETTINGS_ACCESSIBILITY,
	SCR_SELECT_MAP,
	SCR_LAN_HOST_GAME,
	SCR_LAN_JOIN_GAME,
	SCR_LAN_ENTER_JOIN_CODE,
	SCR_EDITOR,
	SCR_IN_GAME,
	SCR_ERROR
} GameScreen;

typedef enum {
	TS_MOVE,
	TS_ATTACK,
	TS_COUNT
} TickStage;

typedef struct {
	GameKind kind;
	GameStatus status;
	GameStatus prev_status; // previous status before user tried to quit 
	v2 grid_base; // coordinate of lower-left corner (not vertex) of hexagon (0, 0) 
	float hex_radius; // radius of each hexagon (dist from center to vertex) 
	u16 width, height; // width and height of hexagonal grid 
	u16 starting_points; // how many points you start the turn with 
	i16 starting_points_increment; // how much the above value is increased/decreased each turn
	u32 points_left; // how many points are left to spend? 
	/* 
		For hotseat games, this is whose turn it is now.
		For network games, this is the player we are controlling on this machine
	*/
	u8 controlled_player;
#if DEBUG
	bool number_pions;
#endif
	float tick_time; // amount of time we've spent on this tick 
	ENUM(TickStage, u8) tick_stage;
	HexPos prev_mouse_hex; // mouse hex last frame 
	Action current_action; // current action we are in the process of doing 
	u8 nplayers;
	u32 npions;
	u32 nactions[PLAYERS_MAX]; // number of actions taken so far 
	Pion pions[PIONS_MAX];
	u8 tiles[GRID_AREA_MAX]; // see tile types above 
	ActionArray actions[PLAYERS_MAX];
} Game;

typedef struct {
	Checkbox fullscreen_cbox;
	Checkbox previews_cbox;
	Checkbox label_pions_cbox;
} SettingsMenu;

enum {
	TEXT_SMALL,
	TEXT_MEDIUM,
	TEXT_BIG
};

typedef struct {
	bool fullscreen;
	bool label_pions;
	bool show_map_previews; 
	u8 text_size;
} Settings;

typedef enum {
	ENT_NONE,
	ENT_TILE,
	ENT_PION
} EntityKind;

// these are the dialogs which can be opened in the map editor
typedef enum {
	EDIALOG_NONE,
	EDIALOG_SAVE_AS,
	EDIALOG_OPEN,
	EDIALOG_WARN_NEW, // are you sure you want to create a new map? there are unsaved changes.
	EDIALOG_WARN_OVERWRITE // are you sure you want to overwrite <map_name>?
} EditorDialog;

typedef struct {
	EntityKind placing;
	// information about the pion we're placing 
	struct {
		u8 player;
	} pion;
	// information about the tile we're placing 
	struct {
		u8 kind;
	} tile;
	EditorDialog dialog;
	TextInput filename_input;
	TextInput starting_points_input, points_inc_input;
	bool unsaved_changes; // have there been changes since the map was last saved?
	char overwritten_map_name[MAP_NAME_BUFSZ]; // name of the map we're about to overwrite (when dialog == EDIALOG_WARN_OVERWRITE)
	char map_name[MAP_NAME_BUFSZ]; // name of the map we're editing 
} Editor;

typedef struct {
	i32 scroll;
	bool scrolling;
} MapSelector;

typedef struct {
	bool initialized;

	int win_width, win_height;
	GameScreen screen;
	v2 mouse_pos;
	bool shift_key_down, control_key_down;
	uint mouse_mask;
	float dt;
	float width_gl;
	ulong frame_number;

#if DEBUG
	bool show_guides; // show horizontal and vertical lines where cursor is 
#endif
	bool trying_to_quit_to_main_menu;

	bool show_back_button;
	char const *back_button_text; // if NULL, defaults to "Back"

	Server *server; // only non-NULL if acting as a server 
	Client *client; // only non-NULL if acting as a client 
	ServerInfo server_info;
	u16 server_password; // set either client: when the user types it in (through the join code) or server: when it's generated 
	Communication comm;
	PlatformProcs procs; // set to frame->procs 
	NetworkStage net_connect_stage;
	TextInput lan_join_code;

	TextInput select_map_input;
	
	Texture spritesheet;
	TextFont main_font, big_font;
	GLProcs gl;
	HexProgram hex_program;
	TitleProgram title_program;
	Settings settings;
	SettingsMenu settings_menu;
	Game game;
	Editor editor;
	SelectMenu main_menu;
	SelectMenu settings_selmenu;
	MapSelector mapsel;

	u32 tmp_mem_used; // this is not measured in bytes, but in MaxAligns 

	float big_fading_message_time;
	v4 big_fading_message_color;
	char big_fading_message[256];
	Sprite sprites[2048];
	u8 font_data[1UL<<20];
	u32 img_mem[1UL<<20];
	MaxAlign tmp_mem[TMP_MEM_BYTES / sizeof(MaxAlign)];


	// stuff for SCR_ERROR 
	char error_message[2048]; 
	bool is_error_bug;
	GameScreen return_to_screen;

#if DEBUG
	u32 magic_number;
	#define MAGIC_NUMBER 0xABCDEFAB
#endif
} State;

// FUNCTION DECLARATIONS 
static void start_turn(Game *game);
