// @TODO(eventually): try sending a whole bunch of malicious data
#include "game.h"
#include <GL/gl.h>
#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <ctype.h>
#if DEBUG
#include <time.h>
#include <sys/stat.h>
#endif
#if PROFILE
#include <time.h>
#ifdef __unix__
static void get_time(struct timespec *ts) {
	clock_gettime(CLOCK_MONOTONIC, ts);
}
#elif defined _WIN32
#include <sysinfoapi.h>
static void get_time(struct timespec *ts) {
	FILETIME ft;
	ULARGE_INTEGER bigint;
	ULONGLONG t;
	GetSystemTimeAsFileTime(&ft);
	bigint.LowPart = ft.dwLowDateTime;
	bigint.HighPart = ft.dwHighDateTime;
	t = bigint.QuadPart;
	ts->tv_sec = t / 10000000;
	ts->tv_nsec = (t % 10000000) * 100;
}
#else
#error "You can't do a profile build because you have an unrecognized operating system."
#endif
#define PROFILE_TIME_PREPARE(name) struct timespec name##_start, name##_end; double name = 0;
#define PROFILE_TIME_BEGIN(name) get_time(&name##_start)
#define PROFILE_TIME_START(name) PROFILE_TIME_PREPARE(name); PROFILE_TIME_BEGIN(name)
#define PROFILE_TIME_END(name) get_time(&name##_end); name = (double)(name##_end.tv_sec - name##_start.tv_sec) * 1000.0 + (double)(name##_end.tv_nsec - name##_start.tv_nsec) / 1000000.0
#else
#define PROFILE_TIME_PREPARE(name) (void)0
#define PROFILE_TIME_BEGIN(name) (void)0
#define PROFILE_TIME_START(name) (void)0
#define PROFILE_TIME_END(name) (void)0
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define STB_TRUETYPE_IMPLEMENTATION
#define STBTT_STATIC
no_warn_start
#include "lib/stb_truetype.h"
no_warn_end

#define MATH_GL
#include "math.c"
#include "decls.h"
#include "colors.c"
#include "util.c"
#include "state.c"
#include "hex.c"
#include "actions.c"
#include "pions.c"
#include "render.c"
#include "buf.c"
#include "comm.c"
#include "ui.c"
#include "settings.c"
#include "maps.c"
#include "editor.c"

// if you run this function, you have to increment tick_stage yourself 
// returns true if anything happened 
static bool do_tick_stage(State *state) {
	Game *game = &state->game;
	bool anything_happened = false;
	ActionArray *actions = game->actions;
	u32 *nactions = game->nactions;
	uint p, nplayers = game->nplayers;
	Pion *pion, *pion_end;
	size_t grid_area = game->width * (size_t)game->height;
	u32 mark = tmp_push(state);

	switch (game->tick_stage) { 
	case TS_MOVE: { // handle move actions 
		bool *block = tmp_alloc_arr(state, bool, grid_area); // these places are blocked because two pions tried to go there at the same time 
		debug_log("MOVE\n");
		/*
			we need to make sure that no two pions go to the same hex. this is more
			complicated than it might seem, because you can get stuff like this:
			a -> b -> [ ] <- c (a, b, and c all don't move)
			a -> b -> [ ] [ ] <- c (a, b, and c all move)
			To do this correctly, we keep track of each pion's "new position" (new_pos), i.e. where it will
			go during this tick. We begin by setting the new_pos of each pion which will move, and if there are
			conflicts (if two have the same new_pos), we push all necessary pions' new_pos's back to their original
			position, and mark the hex where the conflict was with the 'block' array.
			Then, we deal with the stationary pions, pushing back any pions who are trying to move to where
			they are.
		*/
		for (pion = game->pions, pion_end = pion + game->npions; pion != pion_end; ++pion) {
			// the reason we do this at the start, rather than setting new_pos to pos, is so that a -> b -> [ ] works (a shouldn't be blocked by b) 
			pion->new_pos = HEX_POS_INVALID;
		}

		for (p = 0; p < nplayers; ++p) {
			Action *act, *end;
			for (act = actions[p], end = act + nactions[p]; act != end; ++act) {
				if (act->kind == ACT_MOVE) {
					// @OPTIM 
					pion = pion_with_pos(game, act->hex1);
					if (pion && pion->player == p) {
						Pion *other = pion_with_new_pos(game, act->hex2); // @OPTIM 
						if (other) {
							// another pion's here 
							pion->new_pos = pion->pos;
						#if 0
							debug_log("Forcing %p and %p back\n",pion,other);
						#endif
							if (!hex_eq(other->pos, other->new_pos)) {
								// force the other pion to move back 
								pion_force_move_back(game, other);
							}
							block[hex_idx(game, act->hex2)] = true;
						} else if (block[hex_idx(game, act->hex2)]) {
							// 2 pions have already had a conflict here 
						} else {
							/*
								we don't set anything_happened to true here, because this move will be undone
								if we find out about a conflict
							*/
							pion->new_pos = act->hex2;
						}
						// remove this action 
						if (--nactions[p]) {
							// little trick for replacing this action with the last one 
							*act-- = *--end;
						}
					}
				}
			}
		}
		
		for (pion = game->pions, pion_end = pion + game->npions; pion != pion_end; ++pion) {
			if (hex_eq(pion->new_pos, HEX_POS_INVALID)) {
				// this pion is not moving 
				Pion *other = pion_with_new_pos(game, pion->pos);
				if (other) { // someone is trying to move here! 
					pion_force_move_back(game, other);
				}
				pion->new_pos = pion->pos;
			}
		}

		for (pion = game->pions, pion_end = pion + game->npions; pion != pion_end; ++pion) {
			if (!hex_eq(pion->pos, pion->new_pos)) {
				anything_happened = true;
		#if 0
				debug_log("Pion %p is moving\n", pion);
		#endif
			}
			pion->pos = pion->new_pos;
		}
	} break;
	case TS_ATTACK: { // handle attack actions 
		debug_log("ATTACK\n");
		for (p = 0; p < nplayers; ++p) {
			Action *act, *end;
			for (act = actions[p], end = act + nactions[p]; act != end; ++act) {
				if (act->kind == ACT_ATTACK) {
					HexPos hex1 = act->hex1, hex2 = act->hex2;
					// @OPTIM 
					Pion *pion1 = pion_with_pos(game, hex1);
					Pion *pion2 = pion_with_pos(game, hex2);

					if (pion1 && pion2 && pion1->player == p) {
						pion2->will_die = true;
						anything_happened = true;
						// remove this action 
						if (--nactions[p]) {
							*act-- = *--end;
						}
					}
				}
			}
		}
		// remove all pions who we've marked with will_die 
		for (pion = game->pions, pion_end = pion + game->npions; pion != pion_end; ++pion) {
			if (pion->will_die) {
				if (--game->npions) {
					*pion-- = *--pion_end;
				}
			}
		}
	} break;
	case TS_COUNT: assert(0); break;
	}
	tmp_pop(state, mark);
	return anything_happened;
}

static void start_turn(Game *game) {
	game->status = GS_PLAYING;
	game->points_left = game->starting_points;
}

static void reset_round(Game *game) {
	uint p;
	for (p = 0; p < game->nplayers; ++p) {
		game->nactions[p] = 0;
	}
	if (game->kind == GAME_HOTSEAT) game->controlled_player = 0;
	{
		i32 inc = game->starting_points_increment;
		if (inc < 0 && game->starting_points < (u32)-inc) { // can't go below 0 starting points
			game->starting_points = 0;
		} else if (inc > 0 && game->starting_points > STARTING_POINTS_MAX - inc) { // can't go above STARTING_POINTS_MAX
			game->starting_points = STARTING_POINTS_MAX;
		} else {
			game->starting_points = (u16)((i32)game->starting_points + (i32)inc);
		}
	}
	start_turn(game);
}

// load things which need to be reloaded when we load from a save state 
static void init_volatile(State *state) {
	GLProcs *gl = &state->gl;
	Sprite *sprites = state->sprites;
	Settings *settings = &state->settings;
	PlatformProcs *procs = &state->procs;
	void (*(*get_gl_proc)(char const *))(void) = procs->get_gl_proc;
	
	// get GL proc, and exit if it fails 

#define required_gl_proc(name) if (!(gl->name = (GL ## name)get_gl_proc("gl" #name))) { printf("Couldn't get GL proc: %s.\n", #name); exit(-1); }
#define optional_gl_proc(name) gl->name = (GL ## name)get_gl_proc("gl" #name)
	required_gl_proc(AttachShader);
	required_gl_proc(CompileShader);
	required_gl_proc(CreateProgram);
	required_gl_proc(CreateShader);
	required_gl_proc(DeleteProgram);
	required_gl_proc(DeleteShader);
	required_gl_proc(DeleteTextures);
	optional_gl_proc(GenerateMipmap);
	required_gl_proc(GetAttribLocation);
	required_gl_proc(GetProgramInfoLog);
	required_gl_proc(GetProgramiv);
	required_gl_proc(GetShaderInfoLog);
	required_gl_proc(GetShaderiv);
	required_gl_proc(GetUniformLocation);
	required_gl_proc(LinkProgram);
	required_gl_proc(ShaderSource);
	required_gl_proc(Uniform1f);
	required_gl_proc(UniformMatrix4fv);
	required_gl_proc(UseProgram);
	required_gl_proc(VertexAttrib1f);
	required_gl_proc(VertexAttrib2f);
	required_gl_proc(VertexAttrib3f);
	required_gl_proc(VertexAttrib4f);
#undef optional_gl_proc
#undef required_gl_proc

	{ // load shaders 
		{
			GLuint program = compile_and_link_shaders(state, "assets/hex_v.glsl", "assets/hex_f.glsl");
			HexProgram *prog = &state->hex_program;
			prog->program = program;
			prog->bias = get_attrib_loc(state, program, "v_bias");
			prog->transform = get_uniform_loc(state, program, "transform");
		}
		{
			GLuint program = compile_and_link_shaders(state, "assets/title_v.glsl", "assets/title_f.glsl");
			TitleProgram *prog = &state->title_program;
			prog->program = program;
			prog->x_pos = get_attrib_loc(state, program, "v_x_pos");
			prog->transform = get_uniform_loc(state, program, "transform");
		}
	}

	{ // load sprites 
		Texture *spritesheet = &state->spritesheet;
		load_bitmap(state, "assets/spritesheet.bmp", spritesheet);
		sprite_create(spritesheet, sprites + SPR_PION,         20,  20,  256, 256);
		sprite_create(spritesheet, sprites + SPR_ARROW,        20,  290, 250, 100);
		sprite_create(spritesheet, sprites + SPR_ATTACK,       280, 290, 250, 100);
		sprite_create(spritesheet, sprites + SPR_CBOX,         300, 20,  128, 128);
		sprite_create(spritesheet, sprites + SPR_CBOX_CHECKED, 450, 20,  128, 128);
		sprite_create(spritesheet, sprites + SPR_ARROWHEAD,    300, 160, 128, 128);
	}

	{ // load fonts 
		FILE *fp = fopen("assets/font.ttf", "rb");
		size_t file_size;
		u8 *data = state->font_data;

		fseek(fp, 0, SEEK_END);
		file_size = (size_t)ftell(fp);
		fseek(fp, 0, SEEK_SET);
		if (sizeof state->font_data < file_size) {
			printf("!!! Not enough memory for font.\n");
		} else {
			fread(data, 1, file_size, fp);
			load_font(state, &state->main_font, data, main_text_sizes[settings->text_size]);
			load_font(state, &state->big_font, data, big_text_size);
		}
	}
}

// free stuff which needs to be freed before clearing state 
static void free_volatile(State *state) {
	GLProcs *gl = &state->gl;
	if (!state->initialized) return;
	free_texture(state, &state->spritesheet);
	free_font(state, &state->main_font);
	free_font(state, &state->big_font);
	gl->DeleteProgram(state->hex_program.program);
}

#define WINNER_NONE 255
#define WINNER_TIE  254

static u8 game_winner(Game *game) {
	bool have_a_winner = false;
	u8 winner = 0;
	u8 i, nplayers = game->nplayers;
	u32 npions[PLAYERS_MAX] = {0};
	Pion *p, *end;
	for (p = game->pions, end = p + game->npions; p != end; ++p) {
		++npions[p->player];
	}
	if (game->starting_points == 0 && game->starting_points_increment <= 0) {
		// in games with a negative starting points increment, you can run out of points.
		// the winner is the player with the most pions
		bool tie = false;
		for (i = 0; i < nplayers; ++i)
			if (npions[i] > npions[winner])
				winner = i;
		for (i = 0; i < nplayers; ++i)
			if (i != winner && npions[i] == npions[winner])
				tie = true; // someone else has as many pions as the winner.
		return tie ? WINNER_TIE : winner;
	}

	for (i = 0; i < nplayers; ++i) {
		u32 n = npions[i];
		if (n) {
			if (have_a_winner) {
				// turns out multiple people have pions 
				return WINNER_NONE;
			}
			have_a_winner = true;
			winner = i;
		}
	}
	if (have_a_winner)
		return winner;
	// no one has pions left. it's a tie. 
	return WINNER_TIE;
}

static void game_quit(State *state) {
	PlatformProcs *procs = &state->procs;
	if (state->client) {
		procs->client_destroy(state->client);
		state->client = NULL;
	}
	if (state->server) {
		procs->server_destroy(state->server);
		state->server = NULL;
	}
	state->screen = SCR_MAIN_MENU;
	return;
}

// run the game on this frame 
static void run_game(Frame *frame) {
	State *state = frame_state(frame);
	Game *game = &state->game;
	Settings *settings = &state->settings;
	Action *curr_action = &game->current_action;
	Input *input = &frame->input;
	Event *events = input->events;
	u32 nevents = input->nevents;
	Event *events_end = events + nevents;
	float width_gl = state->width_gl;
	v2 mouse_pos = state->mouse_pos;
	float center_x = width_gl * 0.5f;
	v2 screen_center = V2(center_x, 0.5f);
	float dt = (float)frame->dt;
	
	u8 winner = game_winner(game);
	char winner_text[32] = {0};
	v3 winner_color = {0};

	HexPos mouse_hex_pos = {U16_MAX, U16_MAX};
	
	bool control_key_down = state->control_key_down;
	bool shift_key_down = state->shift_key_down;

	(void)control_key_down; (void)shift_key_down;

	state->back_button_text = "Quit";

	hex_setup_game_grid(state, V2(0, 0), V2(width_gl, 0.9f));

	if (winner != WINNER_NONE) {
		game->status = GS_DONE;
		if (game->kind == GAME_HOTSEAT)
			game->controlled_player = PLAYER_NONE;
		state->big_fading_message[0] = 0;
		if (winner == WINNER_TIE) {
			winner_color = V3(0.6f, 0.6f, 0.6f);
			strcpy(winner_text, "Tie");
		} else {
			if (game->kind == GAME_HOTSEAT) {
				sprintf(winner_text, "Player %u won!", winner+1);
				winner_color = v4_xyz(player_colors[winner]);
			} else {
				if (winner == game->controlled_player) {
					winner_color = V3(0.5f, 1.0f, 0.5f);
					strcpy(winner_text, "You won!");
				} else {
					winner_color = V3(1.0f, 0.5f, 0.5f);
					strcpy(winner_text, "You lost!");
				}
			}
		}
	}

	// input stuff 
	mouse_hex_pos = hex_at_pos(game, mouse_pos);
	switch (curr_action->kind) {
	case ACT_MOVE:
	case ACT_ATTACK:
		if (hex_in_bounds(game, mouse_hex_pos))
			curr_action->hex2 = mouse_hex_pos;
		break;
	case ACT_NONE: break;
	}

	{
		Event *e;
		for (e = events; e != events_end; ++e) {
			if (game->status == GS_PLAYING) {
				switch (e->kind) {
				case EVENT_MOUSE_DOWN: {
					MouseEvent *m = &e->u.mouse;
					uint button = m->button;
					if (button == MBUTTON_LEFT) {
						switch (curr_action->kind) {
						case ACT_NONE: {
							if (hex_can_have_pion(game, mouse_hex_pos)) {
								// check if there's another move action with the same starting hex; if so, remove it (you can't have two move actions with the same starting hex) 
								uint player = game->controlled_player;
								Action *action, *action_end;
								for (action = game->actions[player], action_end = action + game->nactions[player]; action != action_end; ++action) {
									if (action->kind == ACT_MOVE && hex_eq(action->hex1, mouse_hex_pos)) {
										game->points_left += action_cost(game, action);
										if (--game->nactions[player]) { // if there still are any actions after replacing this one, 
											*action-- = *--action_end; // little trick for replacing this action with the last of this player's actions 
										}
									}
								}
								curr_action->kind = ACT_MOVE;
								curr_action->hex1 = mouse_hex_pos;
								curr_action->hex2 = hex_pos(U16_MAX, U16_MAX);
								if (game->points_left < action_cost(game, curr_action)) {
									curr_action->kind = ACT_NONE;
									not_enough_points(state);
								}
							}
						} break;
						default: {
							action_finish_current(state);
						} break;
						}
					} else if (button == MBUTTON_RIGHT) {
						switch (curr_action->kind) {
						case ACT_NONE:
							if (hex_can_have_pion(game, mouse_hex_pos)) {
								curr_action->kind = ACT_ATTACK;
								curr_action->hex1 = mouse_hex_pos;
								curr_action->hex2 = hex_pos(U16_MAX, U16_MAX);
								if (game->points_left < action_cost(game, curr_action)) {
									curr_action->kind = ACT_NONE;
									not_enough_points(state);
								}
							}
							break;
						default:
							action_finish_current(state);
							break;
						}
					} else if (button == MBUTTON_MIDDLE) {
						if (curr_action->kind == ACT_NONE) {
							actions_remove_all_at_hex(game, mouse_hex_pos);
						} else {
							curr_action->kind = ACT_NONE;
						}
					}
				} break;
				case EVENT_MOUSE_UP: {
					// when the user clicks somewhere and releases on another hex, do the action 
					if (!hex_eq(mouse_hex_pos, curr_action->hex1))
						action_finish_current(state);
				} break;
				case EVENT_KEY_DOWN: {
					KeyEvent *k = &e->u.key;
					switch (k->key) {
					case KEY_SPACE:
						curr_action->kind = ACT_NONE; // don't leave a partially-completed action sitting there 
						switch (game->kind) {
						case GAME_HOTSEAT:
							++game->controlled_player;
							if (game->controlled_player >= game->nplayers) {
								game->controlled_player = PLAYER_NONE;
								game->status = GS_SIMULATING_TICKS;
								game->tick_time = 0;
							} else {
								// start next turn 
								start_turn(game);
							}
							break;
						case GAME_LAN: {
							Communication *comm = &state->comm;
							// send READY 
							memcpy(comm->sending, NET_READY, NET_READY_SIZE);
							comm_start_sending(comm, NET_READY_SIZE);
							game->status = GS_WAITING_FOR_OTHER_PLAYERS;
							comm_start_receiving(comm); // wait for other players to send READY 
						} break;
						}
						break;
					case KEY_Z: {
						uint p = game->controlled_player;
						u32 *nactions = game->nactions;
						if (nactions[p]) {
							// refund points 
							game->points_left += action_cost(game, &game->actions[p][nactions[p]-1]);
							// remove last action 
							--nactions[p];
						}
					} break;
					case KEY_R: {
						uint p = game->controlled_player;
						if (control_key_down) {
							// reset actions 
							game->nactions[p] = 0;
							game->points_left = game->starting_points;
						}
					} break;
					#if DEBUG
					case KEY_N:
						game->number_pions = !game->number_pions;
						break;
					#endif
					}
				} break;
				case EVENT_MOUSE_MOVE: {
					MouseMoveEvent *m = &e->u.mmove;
					u8 mask = m->mask;
					v2 pos = pixels_to_gl_coords(state, m->x, m->y);
					HexPos hex = hex_at_pos(game, pos);
					// drag to continue move action 
					if (mask & MBUTTON_MASK_LEFT) {
						if (curr_action->kind == ACT_MOVE) {
							HexPos prev_hex = game->prev_mouse_hex;
							if (hexes_are_adjacent(curr_action->hex1, prev_hex) &&
								hexes_are_adjacent(prev_hex, hex)) {
								// finish old action (curr_action->hex1 => prev_hex)
								curr_action->hex2 = prev_hex;
								action_finish_current(state);
								if (hex_can_have_pion(game, prev_hex)) {
									// start new action (prev_hex => hex)
									curr_action->kind = ACT_MOVE;
									curr_action->hex1 = prev_hex;
									curr_action->hex2 = hex;
									if (action_cost(game, curr_action) > game->points_left) {
										curr_action->kind = ACT_NONE;
										not_enough_points(state);
									}
								}
							}
								
						}
					} else if (mask & MBUTTON_MASK_MIDDLE) {
						actions_remove_all_at_hex(game, mouse_hex_pos);
					}
				} break;
				default: break;
				}
			}
		}
	}

	switch (game->status) {
	case GS_TRYING_TO_QUIT:
	case GS_DONE:
	case GS_PLAYING:
		break;
	case GS_SIMULATING_TICKS:
		game->tick_time += dt;
		if (game->tick_time > 1.0f) {
			bool something_happened = false;
			while (1) {
				uint start_stage = game->tick_stage;
				for (; game->tick_stage < TS_COUNT; ++game->tick_stage) {
					if (do_tick_stage(state)) {
						// something happened 
						something_happened = true;
						++game->tick_stage;
						break; 
					}
				}
				if (game->tick_stage == TS_COUNT) game->tick_stage = 0;
				// something happened! let's see what! 
				if (something_happened) break;
				// nothing happened during a whole tick (so we can go to the next round) 
				if (start_stage == 0) break;
			}
			if (something_happened) {
				// keep going 
				game->tick_time = 0;
			} else {
				// we can stop now 
				game->status = GS_PLAYING;
				reset_round(game);
			}
		}
		break;
	case GS_WAITING_FOR_OTHER_PLAYERS: {
		Communication *comm = &state->comm;
		if (!comm->is_receiving && !comm->is_sending) {
			// ready signal has been sent to others & others have sent us a ready signal 

			if (comm->recv_size == NET_READY_SIZE && memcmp(comm->received, NET_READY, NET_READY_SIZE) == 0) {
				// okay, we're all ready. send our actions to the client and receive their actions 
				WriteBuffer b = comm_mkwbuffer(comm);
				u32 nactions = game->nactions[game->controlled_player];
				debug_log("Received ready signal. Sending %lu actions...\n", (ulong)nactions);
				buf_write_actions(&b, game->actions[game->controlled_player], nactions);
				if (b.error) {
					display_error(state, SCR_MAIN_MENU, true, "Error writing actions: %s", b.error_message);
					break;
				}
				comm_start_sending_wbuf(comm, &b);
				comm_start_receiving(comm);
				game->status = GS_TRANSFERRING_ACTIONS;
			} else {
				debug_log("!!! Mysterious message of size %lu\n", (ulong)comm->recv_size);
			}
		}
	} break;
	case GS_TRANSFERRING_ACTIONS: {
		Communication *comm = &state->comm;
		if (!comm->is_receiving && !comm->is_sending) {
			// done sending & receiving actions 
			ReadBuffer r = comm_mkrbuffer(comm);
			game->nactions[!game->controlled_player] = buf_read_actions(&r, game, game->actions[!game->controlled_player]);
			if (r.error) {
				display_error(state, SCR_MAIN_MENU, true, "Error reading actions: %s", r.error_message);
				break;
			}
			game->status = GS_SIMULATING_TICKS;
			game->tick_time = 0;
		}
	} break;
	}
	
	render_game_board(state, true, 1);

	sprites_begin(state);

#if 1
	{
		Action *act, *end;
		uint player = game->controlled_player;
		GameStatus st = game->status == GS_TRYING_TO_QUIT ? game->prev_status : game->status;
		if (st == GS_SIMULATING_TICKS) {
			// show all actions! 
			for (player = 0; player < game->nplayers; ++player) {
				for (act = game->actions[player], end = act + game->nactions[player]; act != end; ++act)
					render_action(state, act, player_colors[player]);
			}
		} else if (st == GS_PLAYING) {
			v4 action_color = V4(1.0f, 1.0f, 1.0f, 0.9f);
			assert(player != PLAYER_NONE);
			// draw action we are planning right now 
			render_action(state, curr_action, action_color);

			action_color.w = 0.7f;
			// draw ghost (planned) actions 
			for (act = game->actions[player], end = act + game->nactions[player]; act != end; ++act)
				render_action(state, act, action_color);
		} else if (st != GS_DONE && player != PLAYER_NONE) {
			// show planned actions 
			for (act = game->actions[player], end = act + game->nactions[player]; act != end; ++act)
				render_action(state, act, player_colors[player]);
		}
	}
#endif
	sprites_end();

	if (game->status == GS_DONE) {
		// banner around text saying who won 
		float x1 = 0, y1 = 0.4f, x2 = width_gl, y2 = 0.6f;
		float m = 0.1f;
		glBegin(GL_QUADS);
		glColor4f(m*winner_color.x, m*winner_color.y, m*winner_color.z, 0.9f);
		glVertex2f(x1,y1);
		glVertex2f(x2,y1);
		glVertex2f(x2,y2);
		glVertex2f(x1,y2);
		glEnd();
	}

	{ // draw small text 
		TextFont *font = &state->main_font;
		font_begin(font);
		
		if (
		#if DEBUG
		game->number_pions || 
		#endif
		settings->label_pions
		) { // pion numbers 
			Pion *pion, *end;
			glColor4f(0,0,0,0.8f);
			for (pion = game->pions, end = pion + game->npions; pion != end; ++pion) {
				char text[32] = {0};
				v2 c = hex_center_pos(game, pion->pos);
				if (settings->label_pions)
					sprintf(text, "%u", pion->player);
				#if DEBUG
				if (game->number_pions) {
					uint id = (uint)(pion - game->pions);
					sprintf(text + strlen(text), "%u", id);
				}
				#endif
				text_render_centered(font, text, c);
			}
		}

		if (game->status == GS_DONE) {
			v3_gl_color(winner_color);
			text_render_centered(font, "Press ESC to return to the main menu", V2(center_x, 0.43f));
		}

		font_end();
	}
	{ // draw big text 
		TextFont *font = &state->big_font;
		font_begin(font);
		if (game->status == GS_PLAYING) { // draw point count 
			char text[32] = {0};
			v2 text_size;
			sprintf(text, U32_FMT, game->points_left);
			text_size = get_text_size(font, text);
			v4_gl_color(points_color);
			text_render(font, text, v2_sub(V2(width_gl-0.05f,0.95f), text_size));
		} else if (game->status == GS_DONE) {
			v3_gl_color(winner_color);
			text_render_centered(font, winner_text, screen_center);
		}
		font_end();
	}

	{
		float size = 0.08f;
		GameStatus st = game->status == GS_TRYING_TO_QUIT ? game->prev_status : game->status;
		glBegin(GL_QUADS);
		// draw current player 
		if (st == GS_PLAYING || (st == GS_DONE && game->kind != GAME_HOTSEAT)) {
			v4_gl_color(player_colors[game->controlled_player]);
		} else {
			glColor3f(0.5f, 0.5f, 0.5f);
		}
		glVertex2f(0.01f, 0.99f);
		glVertex2f(0.01f+size, 0.99f);
		glVertex2f(0.01f+size, 0.99f-size);
		glVertex2f(0.01f, 0.99f-size);
		glEnd();
	}


	
	game->prev_mouse_hex = mouse_hex_pos;
}

static void game_lan_host(State *state) {
	PlatformProcs *procs = &state->procs;
	Server *server = procs->server_create(&state->server_info);
	state->server_password = state->server_info.password;
	if (server) {
		state->server = server;
		state->screen = SCR_LAN_HOST_GAME;
		state->net_connect_stage = NET_CONNECTING;
	} else {
		display_error(state, SCR_MAIN_MENU, false, "%s", procs->comm_geterr());
	}
}

static void game_lan_join(State *state, ServerInfo *server_info) {
	PlatformProcs *procs = &state->procs;
	Client *client = procs->client_create(server_info);
	state->server_password = server_info->password;
	if (client) {
		state->client = client;
		state->screen = SCR_LAN_JOIN_GAME;
		state->net_connect_stage = NET_CONNECTING;
	} else {
		display_error(state, SCR_MAIN_MENU, false, "%s", procs->comm_geterr());
	}
}

static void select_map(State *state) {
	state->screen = SCR_SELECT_MAP;
	text_input_clear(&state->select_map_input);
	state->select_map_input.focus = true;
}

static void go_back(Frame *frame) {
	State *state = frame_state(frame);
	PlatformProcs *procs = &state->procs;
	state->big_fading_message[0] = '\0'; // get rid of big fading message
	if (state->trying_to_quit_to_main_menu) {
		if (state->screen == SCR_IN_GAME) {
			game_quit(state);
		}
		state->screen = SCR_MAIN_MENU;
		state->trying_to_quit_to_main_menu = false;
		return;
	}
	switch (state->screen) {
	case SCR_ERROR:
		state->screen = state->return_to_screen;
		break;
	case SCR_LAN_ENTER_JOIN_CODE:
		state->screen = SCR_MAIN_MENU;
		break;
	case SCR_MAIN_MENU:
		frame->close = true;
		return;
	case SCR_SETTINGS_MENU:
		state->screen = SCR_MAIN_MENU;
		break;
	case SCR_SETTINGS_VIDEO:
	case SCR_SETTINGS_ACCESSIBILITY:
		state->screen = SCR_SETTINGS_MENU;
		break;
	case SCR_LAN_HOST_GAME:
		procs->server_destroy(state->server);
		state->server = NULL;
		state->screen = SCR_MAIN_MENU;
		break;
	case SCR_LAN_JOIN_GAME:
		procs->client_destroy(state->client);
		state->client = NULL;
		state->screen = SCR_MAIN_MENU;
		break;
	case SCR_SELECT_MAP:
		state->screen = SCR_MAIN_MENU;
		break;
	case SCR_IN_GAME: 
		if (game_winner(&state->game) == WINNER_NONE) {
			state->trying_to_quit_to_main_menu = true;
		} else {
			if (state->screen == SCR_IN_GAME) {
				game_quit(state);
			}
			state->screen = SCR_MAIN_MENU;
			state->trying_to_quit_to_main_menu = false;
		}
		break;
	case SCR_EDITOR: {
		Editor *editor = &state->editor;
		if (editor->dialog) {
			// close dialog 
			editor->dialog = EDIALOG_NONE;
			editor->placing = ENT_NONE;
		} else if (editor->placing != ENT_NONE) {
			// stop placing stuff 
			editor->placing = ENT_NONE;
		} else if (editor->unsaved_changes) {
			state->trying_to_quit_to_main_menu = true;
		} else {
			state->screen = SCR_MAIN_MENU;
			state->trying_to_quit_to_main_menu = false;
		}
	} break;
	}
}

#if !DEBUG
static // don't make this static if we're compiling in debug mode, because we might be importing this from a dynamic library 
#endif
void game_frame(Frame *frame) {
	int win_width = frame->width, win_height = frame->height;
	float dt = (float)frame->dt;
	Input *input = &frame->input;
	PlatformProcs *procs = &frame->procs;
	Event *events = input->events;
	u32 nevents = input->nevents;
	Event *events_end = events + nevents;
	State *state = NULL;
	bool *keys_down = input->keys_down;
	bool control_key_down = keys_down[KEY_LCTRL] || keys_down[KEY_RCTRL];
	bool shift_key_down = keys_down[KEY_LSHIFT] || keys_down[KEY_RSHIFT];
	float width_gl = (float)frame->width / (float)frame->height;
	float center_x = width_gl * 0.5f;
	v2 mouse_pos = {0};
	uint mouse_mask = input->mouse_mask;
	GLProcs *gl = NULL;
	Settings *settings = NULL;

	PROFILE_TIME_PREPARE(frame_time);


	(void)mouse_mask; (void)dt;
	
	{
		float bg = 0.0f;
		glClearColor(bg, bg, bg, 1);
	}
	glClear(GL_COLOR_BUFFER_BIT);
	glViewport(0, 0, win_width, win_height);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glOrtho(0, width_gl, 0, 1.0, -1, +1);

	PROFILE_TIME_BEGIN(frame_time);

	(void)control_key_down; (void)shift_key_down;

	if (frame->memory_size < sizeof *state) {
		printf("Didn't get enough memory.\n");
		exit(-1);
	}
	state = frame_state(frame);
	assert(state);

	#if DEBUG
	if (state->initialized && state->magic_number != MAGIC_NUMBER) {
		debug_log("Magic number is wrong. ");
		memset(state, 0, sizeof *state);
	}
	#endif

	state->show_back_button = true; // default to showing the back button
	state->back_button_text = NULL;

	gl = &state->gl;
	settings = &state->settings;

	state->win_width = frame->width;
	state->win_height = frame->height;
	state->mouse_pos = mouse_pos = pixels_to_gl_coords(state, input->mouse_x, input->mouse_y);
	state->shift_key_down = shift_key_down; state->control_key_down = control_key_down;
	state->width_gl = width_gl;
	state->dt = (float)frame->dt;
	state->mouse_pos = mouse_pos;
	state->mouse_mask = mouse_mask;
	state->procs = frame->procs;
	state->tmp_mem_used = 0;
	

	#if DEBUG || PROFILE
	{ // keys you can press to test things 
		Event *e;
		for (e = events; e != events_end; ++e) {
			if (e->kind == EVENT_KEY_DOWN) {
				KeyEvent *k = &e->u.key;
				u16 key = k->key;
				switch (key) {
				case KEY_F5:
					free_volatile(state);
					// reset state 
					memset(state, 0, sizeof *state);
					break;
				case KEY_F6:
					state->game.points_left = U32_MAX;
					break;
				#if DEBUG
				case KEY_F7:
					state->show_guides = !state->show_guides;
					break;
				default:
					if (key >= KEY_F1 && key <= KEY_F4) {
						int num = key - KEY_F1 + 1;
						char filename[32] = {0};
						FILE *fp;
						sprintf(filename, "obj/state%d", num);
						if (control_key_down) {
							fp = fopen(filename, "wb");
							if (fp) {
								fwrite(state, sizeof *state, 1, fp);
								printf("Saved state %d\n", num);
								fclose(fp);
							} else {
								printf("Couldn't open state file: %s\n", filename);
							}
						} else {
							fp = fopen(filename, "rb");
							if (fp) {
								u32 magic = 0;
								fseek(fp, offsetof(State, magic_number), SEEK_SET);
								fread(&magic, sizeof magic, 1, fp);
								if (magic == MAGIC_NUMBER) {
									fseek(fp, 0L, SEEK_SET);
									free_volatile(state);
									fread(state, sizeof *state, 1, fp);
									printf("Loaded state %d\n", num);
									init_volatile(state);
								} else {
									printf("State is corrupted (expected %lx for magic number, but got %lx)\n",
										(ulong)MAGIC_NUMBER, (ulong)magic);
								}
								fclose(fp);
							} else {
								printf("Couldn't load state: %s\n", filename);
							}
						}
					}
				#endif
				}
			}
		}
	}
	#endif

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (!state->initialized) {
		PROFILE_TIME_START(init_time);

	#if DEBUG || PROFILE
		printf("Initializing (state size %luMB/%luMB)...\n", (ulong)sizeof *state >> 20, (ulong)frame->memory_size >> 20);
		fflush(stdout);
	#endif
	
		settings_read(state);
		init_volatile(state);
		frame->fullscreen = settings->fullscreen;

		{
			Communication *comm = &state->comm;
			comm->sending = comm->send_data + 4;
		}
	
		copy_starting_maps_if_needed(state);

		state->initialized = true;
	#if DEBUG
		state->magic_number = MAGIC_NUMBER;
	#endif

	#if PROFILE
		PROFILE_TIME_END(init_time);
		printf("Finished initializing in %.1fms.\n", init_time);
	#endif
	}
#if DEBUG
	bitmap_check_for_changes(state, &state->spritesheet);
#endif

	update_font_if_necessary(state, &state->main_font);
	update_font_if_necessary(state, &state->big_font);
	
	{
		Communication *comm = &state->comm;
		Server *server = state->server;
		Client *client = state->client;
		if (comm->is_receiving) {
			uint nbytes_of_size_received = comm->nbytes_of_size_received;
			{ // receive size 
				u8 *recv_size_buf = comm->recv_size_buf;
				long n = 0;
				if (server) {
					n = procs->server_recv(server, recv_size_buf + nbytes_of_size_received, 4-nbytes_of_size_received);
				} else if (client) {
					n = procs->client_recv(client, recv_size_buf + nbytes_of_size_received, 4-nbytes_of_size_received);
				} else {
					comm->recv_size = 0;
					comm_stop_receiving(comm);
					n = 0;
				}
				if (n < 0) {
					display_error(state, SCR_MAIN_MENU, false, "Error receiving data over the internet: %s.", procs->comm_geterr());
					comm_stop_receiving(comm);
					comm->recv_size = 0;
				} else if (n > 0) {
					nbytes_of_size_received += (uint)n;
					comm->nbytes_of_size_received = (u8)nbytes_of_size_received; 
					if (nbytes_of_size_received == 4) {
						// construct recv_size 
						comm->recv_size = (u32)recv_size_buf[0] + ((u32)recv_size_buf[1] << 8)
							+ (u32)(recv_size_buf[2] << 16) + (u32)(recv_size_buf[3] << 24);
						comm->nrecvd = 0;
						if (comm->recv_size > COMM_MAX_BYTES) {
							display_error(state, SCR_MAIN_MENU, true, "Error receiving data: size of %lu exceeds maximum receive size of " STRINGIFY(COMM_MAX_BYTES),
								(ulong)comm->recv_size);
							comm->recv_size = 0;
							comm_stop_receiving(comm);
						}
					}
				}
			}

			if (nbytes_of_size_received >= 4) {
				long n = 0;
				u8 *recvd = comm->received;
				if (server) {
					n = procs->server_recv(server, recvd + comm->nrecvd, comm->recv_size - comm->nrecvd);
				} else if (client) {
					n = procs->client_recv(client, recvd + comm->nrecvd, comm->recv_size - comm->nrecvd);
				} else { 
					comm->recv_size = 0;
					comm_stop_receiving(comm);
					n = 0;
				}
				if (n < 0) {
					display_error(state, SCR_MAIN_MENU, false, "Error receiving data over the internet: %s.", procs->comm_geterr());
					comm->recv_size = 0;
					comm_stop_receiving(comm);
				} else if (n > 0) {
					comm->nrecvd += (size_t)n;
					if (comm->nrecvd >= comm->recv_size) {
						assert(comm->nrecvd == comm->recv_size);
						comm_stop_receiving(comm);
					}
				}
			}
		}
		if (comm->is_sending) { 
			if (comm->send_size > COMM_MAX_BYTES) {
				display_error(state, SCR_MAIN_MENU, true, "Error sending data: size of %lu exceeds maximum size of " STRINGIFY(COMM_MAX_BYTES) "\n",
					(ulong)comm->send_size);
				comm_stop_sending(comm);
			} else {
				long n = 0;
				size_t send_size = comm->send_size + 4; // total number of bytes to send = 4 bytes for the size + size of data 
				u8 *send_data = comm->send_data;
				if (server) {
					n = procs->server_send(server, send_data, send_size - comm->nsent);
				} else if (client) {
					n = procs->client_send(client, send_data, send_size - comm->nsent);
				} else {
					n = 0;
					comm_stop_sending(comm);
				}
				if (n < 0) {
					display_error(state, SCR_MAIN_MENU, false, "Error sending data over the internet: %s.", procs->comm_geterr());
					comm_stop_sending(comm);
				} else if (n > 0) {
					comm->nsent += (size_t)n;
					if (comm->nsent >= send_size) {
						comm_stop_sending(comm);
					}
				}
			}
		}
	}

	{ // universal events 
		Event *e;
		for (e = events; e != events_end; ++e) {
			switch (e->kind) {
			case EVENT_KEY_DOWN: {
				KeyEvent *key = &e->u.key;
				if (key->key != KEY_ESCAPE) {
					state->trying_to_quit_to_main_menu = false;
				}
				switch (key->key) {
				case KEY_ESCAPE: // deal with the escape key 
					go_back(frame);
					break;
				case KEY_F:
					if (control_key_down)
						frame->fullscreen = !frame->fullscreen;
					break;
				} 
			} break;
			/* 
			on some machines it's not easy to middle/right click.
			so we interpret Shift+Left click as Right click
			and Ctrl+Left click as Middle click
			*/
			case EVENT_MOUSE_DOWN: {
				u8 *button = &e->u.mouse.button;
				if (*button == MBUTTON_LEFT) {
					if (shift_key_down)
						*button = MBUTTON_RIGHT;
					if (control_key_down)
						*button = MBUTTON_MIDDLE;
				}
			} break;
			case EVENT_MOUSE_MOVE: {
				u8 *mask = &e->u.mmove.mask;
				if (*mask & MBUTTON_MASK_LEFT) {
					if (shift_key_down) {
						*mask &= (u8)~MBUTTON_MASK_LEFT;
						*mask |= MBUTTON_MASK_RIGHT;
					}
					if (control_key_down) {
						*mask &= (u8)~MBUTTON_MASK_LEFT;
						*mask |= MBUTTON_MASK_MIDDLE;
					}
				}
			} break;
			}
		}
	}

	// ----RENDERING---- 

#if 0
	{ // debug font rendering 
		TextFont *font = &state->big_font;
		font_begin(font);
		gl_color1f(1);
		// text_render(font, "!", V2(0.5f,0.5f)); 
		glTexCoord2f(0,0); glVertex2f(0,0);
		glTexCoord2f(1,0); glVertex2f(1,0);
		glTexCoord2f(1,1); glVertex2f(1,1);
		glTexCoord2f(0,1); glVertex2f(0,1);
		font_end();

		return;
	}
#endif

	if (state->screen == SCR_MAIN_MENU || state->screen == SCR_SETTINGS_MENU || state->screen == SCR_SETTINGS_VIDEO
		|| state->screen == SCR_SETTINGS_ACCESSIBILITY) {
		// background 
		render_hex_grid(state, 0.1f, V2(-0.1f,-0.1f), 0.2f, false, 30, 30, NULL);
	}
	switch (state->screen) {
	case SCR_MAIN_MENU: {
		SelectMenu *menu = &state->main_menu;
		static char const *const options[] = {
			"Hotseat",
			"Local Network - Host",
			"Local Network - Join",
			"Create a map",
			"Settings",
			"Quit"
		};
		enum {
			HOTSEAT,
			LAN_HOST,
			LAN_JOIN,
			CREATE_MAP,
			SETTINGS,
			QUIT
		};
		int noptions = arr_count(options);
		int chosen_option = -1; // set to something else if an option gets chosen 

		state->show_back_button = false;
		menu->options = options; menu->noptions = noptions;


		{
			Event *e;
			for (e = events; e != events_end; ++e) {
				if (chosen_option == -1)
					chosen_option = selmenu_event(menu, state, e);
			}
		}
			
		switch (chosen_option) {
		case HOTSEAT:
			state->game.kind = GAME_HOTSEAT;
			select_map(state);
			break;
		case LAN_HOST: {
			state->game.kind = GAME_LAN;
			select_map(state);
		} break;
		case LAN_JOIN: {
			state->screen = SCR_LAN_ENTER_JOIN_CODE;
			state->lan_join_code.focus = true;
		} break;
		case CREATE_MAP:
			editor_init(state);
			state->screen = SCR_EDITOR;
			break;
		case SETTINGS:
			state->screen = SCR_SETTINGS_MENU;
			break;
		case QUIT:
			frame->close = true;
			return;
		}

		selmenu_render(menu, state);
		{
			TextFont *font = &state->big_font;
			float pad_y = 0.05f, text_y = 0.9f;
			float y1 = text_y - pad_y, y2 = text_y + pad_y;
			float w = width_gl;
			float border_radius = 1.0f/(float)win_height;
			m4 transform = state_transform(state);
			TitleProgram *prog = &state->title_program;
			GLuint x_pos = prog->x_pos;

			
			gl->UseProgram(prog->program);
			gl->UniformMatrix4fv(prog->transform, 1, GL_FALSE, transform.e);

			glBegin(GL_QUADS);
			gl_color1f(1);
			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y1-border_radius);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y1-border_radius);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y1+border_radius);
			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y1+border_radius);

			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y2-border_radius);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y2-border_radius);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y2+border_radius);
			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y2+border_radius);
			
			{
				v4 c = player_colors[0];
				glColor4f(c.x, c.y, c.z, 1);
			}
			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y1);
			gl->VertexAttrib1f(x_pos, 0); glVertex2f(0, y2);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y2);
			gl->VertexAttrib1f(x_pos, 1); glVertex2f(w, y1);
			glEnd();
			gl->UseProgram(0);

			font_begin(font);
			glColor3f(0, 0, 0);
			text_render_centered(font, "PIONS", V2(center_x, 0.9f));
			font_end();
		}
	} break;

	case SCR_SETTINGS_MENU: {
		SelectMenu *menu = &state->settings_selmenu;
		static char const* const options[] = {
			"Accessibility",
			"Video"
		};
		enum {
			ACCESSIBILITY,
			VIDEO
		};
		int noptions = arr_count(options);
		int chosen_option = -1;

		menu->options = options; menu->noptions = noptions;
		{
			Event *e;
			for (e = events; e != events_end; ++e)
				if (chosen_option == -1)
					chosen_option = selmenu_event(menu, state, e);
		}
		switch (chosen_option) {
		case ACCESSIBILITY:
			state->screen = SCR_SETTINGS_ACCESSIBILITY;
			break;
		case VIDEO:
			state->screen = SCR_SETTINGS_VIDEO;
			break;
		}

		{
			TextFont *font = &state->big_font;
			font_begin(font);
			glColor3f(0.4f, 0.4f, 0.4f);
			text_render_centered(font, "Settings", V2(center_x, 0.9f));
			font_end();
		}
	
		selmenu_render(menu, state);
	} break;

	case SCR_SETTINGS_VIDEO: {
		SettingsMenu *settings_menu = &state->settings_menu;
		TextFont *font = &state->main_font, *big_font = &state->big_font;
		Checkbox *fullscreen = &settings_menu->fullscreen_cbox, *previews = &settings_menu->previews_cbox;
		bool change = false;
		fullscreen->pos = V2(0.01f,0.8f);
		fullscreen->checked = frame->fullscreen;
		previews->pos = V2(0.01f, 0.75f);
		previews->checked = settings->show_map_previews;
		{
			Event *e;
			for (e = events; e != events_end; ++e) {
				change |= checkbox_event(fullscreen, state, e);
				change |= checkbox_event(previews, state, e);
			}
		}
		if (change) {
			frame->fullscreen = settings->fullscreen = fullscreen->checked;
			settings->show_map_previews = previews->checked;
			settings_write(state);
		}
		font_begin(big_font);
		gl_color1f(0.7f);
		text_render_centered(big_font, "Video Settings", V2(center_x, 0.94f));
		font_end();
		font_begin(font);
		checkbox_render_lbl(fullscreen, state, "Fullscreen");
		checkbox_render_lbl(previews, state, "Show map previews");
		font_end();
		sprites_begin(state);
		checkbox_render(fullscreen, state);
		checkbox_render(previews, state);
		sprites_end();
	} break;
	case SCR_SETTINGS_ACCESSIBILITY: {
		SettingsMenu *settings_menu = &state->settings_menu;
		TextFont *font = &state->main_font, *big_font = &state->big_font;
		Button text_small = {0}, text_medium = {0}, text_big = {0};
		bool change = false;
		Checkbox *lbl_pions = &settings_menu->label_pions_cbox;
		char const *font_size_label_text = "Text size:";
		u8 text_size = settings->text_size;
		text_small.pos = V2(0.2f, 0.7f);    text_small.text  = "Small";  text_small.disabled  = text_size == TEXT_SMALL;
		text_medium.pos = V2(0.325f, 0.7f); text_medium.text = "Medium"; text_medium.disabled = text_size == TEXT_MEDIUM;
		text_big.pos = V2(0.5f, 0.7f);      text_big.text    = "Big";    text_big.disabled    = text_size == TEXT_BIG;
		lbl_pions->pos = V2(0.01f,0.8f);
		lbl_pions->checked = settings->label_pions;

		{
			Event *e;
			for (e = events; e != events_end; ++e) {
				change |= checkbox_event(lbl_pions, state, e);
				if (button_event(&text_small, state, e)) {
					text_size = TEXT_SMALL;
					change = true;
				}
				if (button_event(&text_medium, state, e)) {
					text_size = TEXT_MEDIUM;
					change = true;
				}
				if (button_event(&text_big, state, e)) {
					text_size = TEXT_BIG;
					change = true;
				}
			}
		}

		if (change) {
			settings->label_pions = lbl_pions->checked;
			if (settings->text_size != text_size) {
				settings->text_size = text_size;
				reload_font(state, font, main_text_sizes[text_size]);
			}
			settings_write(state);
		}

		font_begin(big_font);
		gl_color1f(0.7f);
		text_render_centered(big_font, "Accessibility Settings", V2(center_x, 0.94f));
		font_end();
		font_begin(font);
		button_render(&text_small, state);
		button_render(&text_medium, state);
		button_render(&text_big, state);
		checkbox_render_lbl(lbl_pions, state, "Label pion colors");
		text_render(font, font_size_label_text, V2(0.01f, 0.7f));
		font_end();
		sprites_begin(state);
		checkbox_render(lbl_pions, state);
		sprites_end();
	} break;
	case SCR_LAN_HOST_GAME: {
		TextFont *font = &state->big_font;
		Server *server = state->server;
		Communication *comm = &state->comm;
		assert(server);
		switch (state->net_connect_stage) {
		case NET_CONNECTING: {
			ClientInfo info = {0};
			if (procs->server_check_for_client(server, &info)) {
				debug_log("Someone connected!\n");
				comm_start_receiving(comm); // receive password 
				state->net_connect_stage = NET_VERIFYING_CLIENT;
			}
		} break;
		case NET_VERIFYING_CLIENT: {
			if (!comm->is_receiving) {
				ReadBuffer r = comm_mkrbuffer(comm);
				// received password 
				if (buf_read_u16(&r) == state->server_password) {
					u32 v = buf_read_u32(&r);
					if (v == NET_VERSION) {
						Game *game = &state->game;
						debug_log("Client verified successfully.\n");
						{
							WriteBuffer b = comm_mkwbuffer(comm);
							buf_write_game_setup(&b, game, !game->controlled_player);
							if (b.error) {
								display_error(state, SCR_MAIN_MENU, true, "Error sending game: %s", b.error_message);
								procs->server_destroy(server);
								state->server = NULL;
							} else {
								comm_start_sending_wbuf(comm, &b);
								state->net_connect_stage = NET_TRANSFERRING_GAME_DETAILS;
							}
						}
					} else {
						printf("Version mismatch. They've got %u, but we've got %u.", (uint)v, NET_VERSION);
						// let them know 
						{ // will probably go through 
							u32 zero = 0;
							procs->server_send(server, &zero, sizeof zero);
						}
						procs->server_disconnect_from_client(server);
						state->net_connect_stage = NET_CONNECTING;
					}
				} else {
				#if DEBUG
					printf("Got a wrong password.\n");
					// let them know 
					{ // will probably go through 
						u32 zero = 0;
						procs->server_send(server, &zero, sizeof zero);
					}
					procs->server_disconnect_from_client(server);
					state->net_connect_stage = NET_CONNECTING;
				#endif
				}
			}
		} break;
		case NET_TRANSFERRING_GAME_DETAILS: {
			if (!comm->is_sending) {
				// done sending game details 
				state->screen = SCR_IN_GAME;
			}
		} break;
		}
		font_begin(font);
		gl_color_text();
		{
			char text[256] = {0};
			ServerInfo *info = &state->server_info;
			IPv4 ip = info->ip;
			// most of the time, the local IP address is 192.168.0.X, so we can omit the 192.168.0 part 
			if (ip.a == 192 && ip.b == 168 && ip.c == 0) { 
				sprintf(text, "Join code: %02x%01x%04x", ip.d, info->port & 0xF, info->password);
			} else {
				sprintf(text, "Join code: %02x%02x%02x%02x%01x%04x", ip.a, ip.b, ip.c, ip.d, info->port & 0xF, info->password);
			}
			text_render_centered(font, text, V2(center_x, 0.3f));
		}
		text_render_centered(font, "Waiting for someone to join...", V2(width_gl*0.5f,0.7f));
		font_end();
	} break;
	case SCR_LAN_ENTER_JOIN_CODE: {
		char text[256] = {0};
		TextFont *font = &state->big_font;
		TextInput *join_code = &state->lan_join_code;

		join_code->char_predicate = is_lowercase_hex_digit;
		join_code->max_len = 13;
		join_code->pos = V2(center_x, 0.4f);
		join_code->placeholder = "Join code";
		join_code->center = true;
		join_code->font = font;
		join_code->width = 0.6f;

		gl_color_text();
		{
			Event *e;
			for (e = events; e != events_end; ++e) {
				if (text_input_event(join_code, state, e)) {
					char *code = join_code->text;
					ServerInfo info = {0};
					IPv4 *ip = &info.ip;
					info.port = PORT_UPPER_PART << 4;
					if (strlen(code) == 7) {
						uint ip_d, port_lower_4bits, passwd;
						if (sscanf(code, "%02x%01x%04x", &ip_d, &port_lower_4bits, &passwd) != 3) {
							display_error(state, SCR_LAN_ENTER_JOIN_CODE, true, "!!! Can't parse join code! This shouldn't be happening!");
							break;
						}
						info.password = (u16)passwd;
						ip->a = 192;
						ip->b = 168;
						ip->c = 0;
						ip->d = (u8)ip_d;
						info.port |= (u16)port_lower_4bits;
					} else if (strlen(code) == 13) {
						uint ip_a, ip_b, ip_c, ip_d, port_lower_4bits, passwd;
						if (sscanf(code, "%02x%02x%02x%02x%01x%04x", &ip_a, &ip_b, &ip_c, &ip_d, &port_lower_4bits, &passwd) != 6) {
							display_error(state, SCR_LAN_ENTER_JOIN_CODE, true, "Can't parse join code! This shouldn't be happening!");
							break;
						}
						info.password = (u16)passwd;
						ip->a = (u8)ip_a;
						ip->b = (u8)ip_b;
						ip->c = (u8)ip_c;
						ip->d = (u8)ip_d;
						info.port |= (u16)port_lower_4bits;
					} else {
						display_error(state, SCR_LAN_ENTER_JOIN_CODE, false, "Invalid join code. It should be 7 or 13 digits long.");
						break;
					}
					game_lan_join(state, &info);
				}
			}
		}
		font_begin(font);
		text_render_centered(font, "Join a local network game", V2(center_x, 0.55f));
		text_render_centered(font, text, V2(center_x, 0.45f));
		font_end();

		text_input_render(join_code, state);
	} break;
	case SCR_LAN_JOIN_GAME: {
		TextFont *font = &state->big_font;
		Client *client = state->client;
		Communication *comm = &state->comm;
		assert(client);
		switch (state->net_connect_stage) {
		case NET_CONNECTING: {
			int status = procs->client_is_connected(client);
			if (status == -1) {
				display_error(state, SCR_LAN_ENTER_JOIN_CODE, false, "%s", procs->comm_geterr());
				procs->client_destroy(client);
				state->client = NULL;
			} else if (status == 1) {
				WriteBuffer b = comm_mkwbuffer(comm);
				u16 passwd = state->server_password;
				debug_log("Connected! Sending password and version...\n");
				buf_write_u16(&b, passwd);
				buf_write_u32(&b, NET_VERSION);
				comm_start_sending_wbuf(comm, &b);
				state->net_connect_stage = NET_VERIFYING_CLIENT;
			}
		} break;
		case NET_VERIFYING_CLIENT:
			if (!comm->is_sending) {
				// done sending password. now receive game info 
				comm_start_receiving(comm);
				state->net_connect_stage = NET_TRANSFERRING_GAME_DETAILS;
			}
			break;
		case NET_TRANSFERRING_GAME_DETAILS:
			if (!comm->is_receiving) {
				// done receiving game info 
				Game *game = &state->game;
				ReadBuffer b = comm_mkrbuffer(comm);
				if (comm->recv_size == 0) {
					display_error(state, SCR_LAN_ENTER_JOIN_CODE, false, "Wrong join code.");
				} else if (comm->recv_size == 1) {
					display_error(state, SCR_MAIN_MENU, false, "You have incompatible versions of the game.");
				} else {
					buf_read_game_setup(&b, state);
					game->kind = GAME_LAN;
					if (b.error) {
						display_error(state, SCR_MAIN_MENU, true, "Error reading game: %s", b.error_message);
					} else {
						state->screen = SCR_IN_GAME;
					}
				}
			}
			break;
		}
		font_begin(font);
		text_render_centered(font, "Connecting...", V2(center_x,0.5f));
		font_end();
	} break;
	case SCR_SELECT_MAP: {
		TextInput *search = &state->select_map_input;
		{
			search->pos = V2(center_x, 0.85f);
			search->center = true;
			search->char_predicate = can_char_be_used_in_map_filename;
			search->placeholder = "Search...";
		}
		{
			char map_name[MAP_NAME_BUFSZ];
			Game *game = &state->game;
			GameKind game_kind = game->kind; // save this because it will be overwritten if we load a map
			int ret = map_selector(frame, search, 0.1f, 0.75f, 0.8f, map_name);
			if (ret == 0 /* pre-existing map */ && *map_name) {
				int load_status = map_load(state, map_name, true, SCR_SELECT_MAP);
				if (load_status == 1) {
					// all good 
					game->kind = game_kind;
					switch (game->kind) {
					case GAME_LAN:
						game->controlled_player = (u8)(rand() % 2); // randomize starting player
						game->kind = GAME_LAN;
						game_lan_host(state);
						break;
					case GAME_HOTSEAT:
						state->screen = SCR_IN_GAME;
						game->kind = GAME_HOTSEAT;
						game->controlled_player = 0;
						break;
					}
				} else if (load_status == 0) {
					display_error(state, SCR_MAIN_MENU, false, "Map not found.");
				} else if (load_status == -1) {
					// map_load handled this
				}
				text_input_clear(search);
			} else if (*map_name) {
				// if user pressed enter with a non-existent map in the search box, clear input 
				text_input_clear(search);
			}
		}

		{
			TextFont *font = &state->big_font;
			font_begin(font);
			gl_color_text();
			text_render_centered(font, "Select a map", V2(center_x,0.95f));
			font_end();
		}
	} break;
	case SCR_EDITOR: {
		run_game_editor(frame);
	} break;
	case SCR_IN_GAME:
		run_game(frame);
		break;
	case SCR_ERROR: {
		TextFont *big_font = &state->big_font;
		TextFont *font = &state->main_font;

		font_begin(big_font);
		v4_gl_color(error_message_color);
		text_render_centered(big_font, "Error", V2(center_x, 0.55f));
		font_end();

		font_begin(font);
		text_render_centered(font, state->error_message, V2(center_x, 0.45f));
		if (state->is_error_bug) {
			text_render_centered(font, "This shouldn't be happening.", V2(center_x, 0.40f));
			text_render_centered(font, "You can report this problem to " BUG_EMAIL ".", V2(center_x, 0.30f));
		}
		font_end();
	} break;
	}
	if (state->screen == SCR_IN_GAME || state->screen == SCR_EDITOR) { // pion counters
		TextFont *font = &state->main_font;
		font_begin(font);
		{ // pion counters 
			Pion *pion, *end;
			Game *game = &state->game;
			u32 player_npions[PLAYERS_MAX] = {0}; // number of pions for each player 
			u32 min_pions = U32_MAX; // least number of pions a player has 
			uint player;
			v2 text_pos = V2(0.15f, 0.95f);
			for (pion = game->pions, end = pion + game->npions; pion != end; ++pion) {
				++player_npions[pion->player];
			}

			for (player = 0; player < game->nplayers; ++player) {
				u32 n = player_npions[player];
				if (min_pions > n)
					min_pions = n;
			}
			
			for (player = 0; player < game->nplayers; ++player) {
				char buf[32] = {0};
				u32 n = player_npions[player];
				// if one player has more pions than the others, highlight that 
				char const *s = n > min_pions ? "*" : "";
				sprintf(buf, "%u%s", n, s);
				v4_gl_color(player_colors[player]);
				text_render(font, buf, text_pos);
				text_pos.x += get_text_size(font, buf).x + 0.03f;
			}
		}
		font_end();
	}

	{ // big fading message 
		char *message = state->big_fading_message;
		if (*message) { // if there is a message, 
			TextFont *font = &state->big_font;
			float t = state->big_fading_message_time += dt;
			float const fade_time = 2.5f;
			if (t >= fade_time) {
				// get rid of message. it's all faded now 
				*message = 0;
			} else {
				v4 color = state->big_fading_message_color;
				v2 text_size = get_text_size(font, message);
				float progress = t/fade_time;
				color.w *= sqrtf(1-progress);

				glBegin(GL_QUADS);
				color.w *= 0.2f;
				v4_gl_color(color);
				color.w /= 0.2f;
				{
					float dy = text_size.y * 0.9f;
					float y1 = 0.5f+dy, y2 = 0.5f-dy;
					glVertex2f(0.0f, y1);
					glVertex2f(0.0f, y2);
					glVertex2f(width_gl, y2);
					glVertex2f(width_gl, y1);
				}
				glEnd();

				font_begin(font);
				v4_gl_color(color);
				text_render_centered(font, message, V2(center_x, 0.5f));
				font_end();
			}
		}
	}


	if (state->trying_to_quit_to_main_menu) {
		float pad_x = 0.2f, pad_y = 0.2f;
		float x1 = pad_x, y1 = pad_y;
		float x2 = width_gl - pad_x, y2 = 1-pad_y;
		TextFont *big_font = &state->big_font, *font = &state->main_font;

		glBegin(GL_QUADS);
		glColor4f(0.1f,0,0,0.9f);
		glVertex2f(x1,y1);
		glVertex2f(x1,y2);
		glVertex2f(x2,y2);
		glVertex2f(x2,y1);
		glEnd();
		
		
		glColor3f(1.0f,0.0f,0.0f);
		font_begin(big_font);
		text_render_centered(big_font, "Quit?", V2(center_x, 0.55f));
		font_end();
		font_begin(font);
		text_render_centered(font, "Press escape again to quit to main menu.", V2(center_x, 0.45f));
		text_render_centered(font, "Press any other key to cancel.", V2(center_x, 0.4f));
		if (state->screen == SCR_EDITOR) {
			text_render_centered(font, "Any unsaved changes will be discarded.", V2(center_x, 0.35f));
		}
		font_end();
	}

	if (state->show_back_button) { // a simple back button used by most screens 
		Button back_button = {0};
		Event *event;
		back_button.pos = V2(0.01f,0.01f);
		back_button.text = state->back_button_text ? state->back_button_text : "Back";
		font_begin(&state->main_font);
		button_render(&back_button, state);
		font_end();
		for (event = events; event != events_end; ++event) {
			if (button_event(&back_button, state, event))
				go_back(frame);
		}
	}

#if DEBUG
	if (state->show_guides) {
		glBegin(GL_LINES);
		glColor3f(0,1,0);
		glVertex2f(0, mouse_pos.y);
		glVertex2f(width_gl, mouse_pos.y);
		glVertex2f(mouse_pos.x, 0);
		glVertex2f(mouse_pos.x, 1);
		glEnd();
		{
			TextFont *font = &state->main_font;
			font_begin(font);
			{
				char str[64] = {0};
				sprintf(str, "x = %f", mouse_pos.x);
				text_render(font, str, V2(mouse_pos.x+0.01f, 0.01f));
			}
			{
				char str[64] = {0};
				sprintf(str, "y = %f", mouse_pos.y);
				text_render(font, str, V2(0.01f, mouse_pos.y+0.01f));
			}
			font_end();
		}
	}
#endif

#if PROFILE
	PROFILE_TIME_END(frame_time);
	{
		TextFont *font = &state->main_font;
		char text[32] = {0};
		v2 text_size;
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, font->texture);
		sprintf(text, "dt: %.1fms", frame_time);
		text_size = get_text_size(font, text);
		glBegin(GL_QUADS);
		gl_color_text();
		text_render(font, text, V2((width_gl - text_size.x) * 0.5f, 0.99f - text_size.y));
		glEnd();
		glDisable(GL_TEXTURE_2D);
	}
#endif
}
