#ifdef DISABLE_NETWORKING
#include "sockets-none.c"
#else
struct Server {
	SOCKET socket;
	SOCKET conn;
};

struct Client {
	SOCKET socket;
	bool connected;
	struct sockaddr_in addr;
};

static char comm_err[256];

static char *comm_geterr(void) {
	return comm_err;
}

static void comm_seterr(char const *fmt, ...) {
	va_list args;
	va_start(args, fmt);
	vsprintf_s(comm_err, sizeof comm_err, fmt, args);
	va_end(args);
}

static IPv4 sockaddr_to_ipv4(SOCKET_ADDRESS *addr) {
	SOCKADDR_IN *ipv4 = (SOCKADDR_IN *)addr->lpSockaddr; 
	IN_ADDR in = ipv4->sin_addr;
	IPv4 ip = {0};
	ip.a = in.S_un.S_un_b.s_b1;
	ip.b = in.S_un.S_un_b.s_b2;
	ip.c = in.S_un.S_un_b.s_b3;
	ip.d = in.S_un.S_un_b.s_b4;
	return ip;
}

static bool get_ip_address(IPv4 *ip) {
	ULONG sz = 16384;
	bool success = false;
	IP_ADAPTER_ADDRESSES *ip_addrs = malloc(sz);
	if (ip_addrs) {
		// don't fetch information we don't need 
		ULONG flags = GAA_FLAG_SKIP_MULTICAST | GAA_FLAG_SKIP_DNS_SERVER | GAA_FLAG_SKIP_FRIENDLY_NAME;
		DWORD ret = GetAdaptersAddresses(AF_INET, flags, NULL, ip_addrs, &sz);
		if (ret == 0) {
			IP_ADAPTER_ADDRESSES *curr;
			IP_ADAPTER_UNICAST_ADDRESS *unicast_addr;
			for (curr = ip_addrs; curr; curr = curr->Next) {
				if (curr->OperStatus != IfOperStatusUp) continue;
				if (curr->IfType == IF_TYPE_SOFTWARE_LOOPBACK) continue;
				for (unicast_addr = curr->FirstUnicastAddress; unicast_addr; unicast_addr = unicast_addr->Next) {
					SOCKET_ADDRESS *sock_addr = &unicast_addr->Address;
					*ip = sockaddr_to_ipv4(sock_addr);
					success = true;
					break;
				}
			}
		}
		free(ip_addrs);
	}
	return success;
}

static Server *server_create(ServerInfo *info) {
	Server *s = (Server *)calloc(1, sizeof *s);
	bool success = false;
	info->password = (u16)rand();

	if (s) {
		if (get_ip_address(&info->ip)) {
			SOCKET sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
			if (sock != INVALID_SOCKET) {
				ULONG nonblocking = 1;
				// set socket to non-blocking mode 
				if (ioctlsocket(sock, FIONBIO, &nonblocking) == 0) {
					struct sockaddr_in addr = {0};
					int port_lower_part;
					addr.sin_family = AF_INET;
					addr.sin_addr.s_addr = htonl(INADDR_ANY);
					s->socket = sock;
					for (port_lower_part = 0; port_lower_part < 0x10; ++port_lower_part) {
						u16 port = info->port = ((u16)PORT_UPPER_PART << 4) | port_lower_part;
						addr.sin_port = htons(port);
						if (bind(sock, (struct sockaddr const *)&addr, sizeof addr) != SOCKET_ERROR) {
							if (listen(sock, SOMAXCONN) != SOCKET_ERROR) {
								success = true;
								break;
							} else {
								comm_seterr("Couldn't listen on port %u.", port);
								break;
							}
						} else {
							int err = WSAGetLastError();
							if (err == WSAEADDRINUSE) {
								// try next port number 
							} else {
								comm_seterr("Couldn't bind socket to port %u.", port);
								break;
							}
						}
					}
					if (port_lower_part == 0x10) {
						comm_seterr("Couldn't find an available port.");
					}
				} else {
					comm_seterr("Couldn't set socket to non-blocking mode.");
				}
				if (!success) closesocket(sock);
			} else {
				comm_seterr("Couldn't create socket.");
			}
		} else {
			comm_seterr("Couldn't get IP address. Are you connected to the internet?");
		}
		if (!success) free(s);
	} else {
		comm_seterr("Out of memory.");
	}
	return success ? s : NULL;
}

static bool server_check_for_client(Server *server, ClientInfo *info) {
	SOCKET conn = accept(server->socket, NULL, NULL);
	if (conn == INVALID_SOCKET) {
		return false;
	} else {
		server->conn = conn;
		info->unused = 0;
		return true;
	}
}

static long deal_with_send(int ret) {
	if (ret == SOCKET_ERROR) {
		int err = WSAGetLastError();
		if (err == WSAEWOULDBLOCK) {
			return 0; // not serious 
		} else {
			comm_seterr("Error sending data over the internet (error code %d).", err);
			return -1;
		}
	} else {
		return (long)ret;
	}
}

static long deal_with_recv(int ret) {
	if (ret == SOCKET_ERROR) {
		int err = WSAGetLastError();
		if (err == WSAEWOULDBLOCK) {
			return 0; // not serious 
		} else {
			comm_seterr("Error sending data over the internet (error code %d).", err);
			return -1;
		}
	} else {
		return (long)ret;
	}
}

static long server_send(Server *server, void const *data, size_t n) {
	return deal_with_send(send(server->conn, data, (int)n, 0));
}
static long server_recv(Server *server, void *data, size_t n) {
	return deal_with_recv(recv(server->conn, data, (int)n, 0));
}

static void server_disconnect_from_client(Server *server) {
	if (server->conn) closesocket(server->conn);
	server->conn = 0;
}

static void server_destroy(Server *server) {
	if (server->conn) closesocket(server->conn);
	if (server->socket) closesocket(server->socket);
	free(server);
}

static Client *client_create(ServerInfo *info) {
	Client *c = (Client *)calloc(1, sizeof *c);
	bool success = false;
	if (c) {
		SOCKET sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (sock != INVALID_SOCKET) {
			struct sockaddr_in addr = {0};
			ULONG nonblocking = 1;
			c->socket = sock;
			if (ioctlsocket(sock, FIONBIO, &nonblocking) == 0) {
				IPv4 ip = info->ip;
				u16 port = info->port;
				addr.sin_family = AF_INET;
				addr.sin_port = htons(port);
				addr.sin_addr.s_addr = ((u32)ip.d << 24) | ((u32)ip.c << 16) | ((u32)ip.b << 8) | ip.a;
				c->addr = addr;
				if (connect(sock, (struct sockaddr *)&addr, sizeof addr) != SOCKET_ERROR) {
					c->connected = true;
					success = true;
				} else {
					int err = WSAGetLastError();
					if (err == WSAEISCONN) {
						success = true;
						c->connected = true;
					} if (err == WSAEWOULDBLOCK || err == WSAEINPROGRESS) {
						// that's okay. we're not connected yet, but we're trying. 
						success = true;
					} else {
						comm_seterr("Error connecting to server.");
					}
				}
			} else {
				comm_seterr("Error setting socket to non-blocking.");
			}
			if (!success) closesocket(sock);
		} else {
			comm_seterr("Error creating socket.");
		}
		if (!success) free(c);
	} else {
		comm_seterr("Out of memory.");
	}
	return success ? c : NULL;
}

static long client_send(Client *client, void const *data, size_t n) {
	return deal_with_send(send(client->socket, data, (int)n, 0));
}
static long client_recv(Client *client, void *data, size_t n) {
	return deal_with_recv(recv(client->socket, data, (int)n, 0));
}

static int client_is_connected(Client *client) {
	struct sockaddr_in addr = client->addr;
	if (client->connected) return 1;
	if (connect(client->socket, (struct sockaddr *)&addr, sizeof addr) != SOCKET_ERROR) {
		client->connected = true;
		return 1;
	} else {
		int err = WSAGetLastError();
		if (err == WSAEISCONN) { // connected 
			client->connected = true;
			return 1;
		} else if (err == WSAEWOULDBLOCK || err == WSAEINPROGRESS || err == WSAEALREADY) {
			return 0;
		} else {
			comm_seterr("Error connecting to server (code %d).", err);
			return -1;
		}
	}
}

static void client_destroy(Client *c) {
	closesocket(c->socket);
	free(c);
}

#if 0
int main(void) {
	ServerInfo info = {0};
	Server *server;
	
	srand((unsigned)time(NULL));
	{
		WSADATA wsa_data;
		if (WSAStartup(MAKEWORD(2,2), &wsa_data) != 0) {
			printf("Error: WSAStartup failed.\n");
			return 1;
		}
	}

	server = server_create(&info);
	if (server) {
		printf("IP: ");
		print_ip(info.ip);
		printf("\n");
		printf("Password: %04x\n", info.password);
		printf("Port: %04x\n", info.port);
		server_destroy(server);
	} else {
		printf("Error: %s\n", comm_geterr());
	}
	return 0;
}
#endif

#endif // DISABLE_NETWORKING 
